"""
"""
import collections
import bpy
from batoms.butils import object_mode
import numpy as np
from time import time
from pprint import pprint


class Setting():
    """
    Setting object

    The Setting object store the other information for batoms.

    Parameters:

    label: str
        The label define the batoms object that a Setting belong to.
    """
    def __init__(self, label) -> None:
        self.label = label
        self.name = 'base'
    def get_data(self):
        data = {}
        for b in self.collection:
            data[b.name] = b
        return data
    @property
    def collection(self):
        return self.get_collection()
    def get_collection(self):
        collection = getattr(bpy.data.collections[self.label], self.name)
        return collection
    @property
    def species(self):
        return self.get_species()
    def get_species(self) -> dict:
        """
        read species from collection.
        """
        species = {}
        coll_atom = bpy.data.collections['%s_atom'%self.label]
        for ba in coll_atom.objects:
            species[ba.batom.species] = {
                'color': ba.children[0].data.materials[0].diffuse_color,
                'radius': ba.batom.radius}
        return species
    @property
    def data(self):
        return self.get_data()
    def __getitem__(self, index):
        item = self.find(index)
        if item is None:
            raise Exception('%s not in %s setting'%(index, self.name))
        return item
    def __setitem__(self, index, value):
        """
        Add bondpair one by one
        """
        collection = self.collection
        item = self.find(index)
        
    def copy(self, label):
        object_mode()
        bondsetting = self.__class__(label)
        for key, b in self.data.items():
            bondsetting[key] = b.as_list()
        return bondsetting
    def __add__(self, other):
        self += other
        return self
    def __iadd__(self, other):
        self.extend(other)
        return self
    def extend(self, other):
        for key, value in other.data.items():
            self[key] = value.as_list()
        return self
    def __iter__(self):
        item = self.collection
        for i in range(len(item)):
            yield item[i]
    def __len__(self):
        return len(self.collection)
    def find(self, name):
        i = self.collection.find(name)
        if i == -1:
            return None
        else:
            return self.collection[i]


class Bondsetting(Setting):
    """
    Bondsetting object

    The Bondsetting object store the bondpair information.

    Parameters:

    label: str
        The label define the batoms object that a Bondsetting belong to.
    cutoff: float
        Cutoff used to calculate the maxmium bondlength for bond pairs.
    """
    def __init__(self, label, cutoff = 1.3) -> None:
        Setting.__init__(self, label)
        self.label = label
        self.name = 'bbond'
        self.cutoff = cutoff
        self.set_default(self.species, cutoff)
    def __setitem__(self, index, value):
        """
        Add bondpair one by one
        """
        bond = self.find(index)
        if bond is None:
            bond = self.collection.add()
        bond.symbol1 = index.split('-')[0]
        bond.symbol2 = index.split('-')[1]
        bond.name = index
        bond.min = value[0]
        bond.max = value[1]
        bond.search = value[2]
        bond.polyhedra = value[3]
        if len(value) == 8:
            bond.color1 = value[4]
            bond.color2 = value[5]
            bond.width = value[6]
            bond.style = value[7]
    def set_default(self, species, cutoff = 1.3):
        """
        """
        bondtable = get_bondtable(species, cutoff=cutoff)
        for key, value in bondtable.items():
            self['%s-%s'%(key[0], key[1])] = value
    def extend(self, other):
        for key, value in other.data.items():
            self[key] = value.as_list()
        # new
        species1 = set(self.species)
        species2 = set(other.species)
        nspecies1 = species1 - species2
        nspecies2 = species2 - species1
        for sp1 in nspecies1:
            for sp2 in nspecies2:
                species = {sp1: self.species[sp1], sp2: other.species[sp2]}
                self.set_default(species, self.cutoff)
    def add_bonds(self, bondpair):
        for key in bondpair:
            species = {sp: self.species[sp] for sp in key}
            self.set_default(species)
    def remove_bonds(self, bondpair):
        if isinstance(bondpair[0], str):
            bondpair = [bondpair]
        for key in bondpair:
            name = '%s-%s'%(key[0], key[1])
            i = self.collection.find(name)
            if i != -1:
                self.collection.remove(i)
    def __repr__(self) -> str:
        s = '-'*60 + '\n'
        s = 'Bondpair      min     max   Search_bond    Polyhedra style\n'
        for b in self.collection:
            s += '{0:10s} {1:4.3f}   {2:4.3f}      {3:10s}   {4:10s}  {5:4s} \n'.format(\
                b.name, b.min, b.max, str(b.search), str(b.polyhedra), b.style)
        s += '-'*60 + '\n'
        return s
    @property
    def cutoff_dict(self):
        cutoff_dict = {}
        for b in self.collection:
            cutoff_dict[(b.symbol1, b.symbol2)] = [b.min, b.max]
        return cutoff_dict
    @property
    def maxcutoff(self):
        maxcutoff = 2
        for bl in self.cutoff_dict.values():
            if bl[1] > maxcutoff:
                maxcutoff = bl[1]
        return maxcutoff
    def search_bond_list(self, atoms, bondlists0, offsets_skin0):
        bondlist1 = []
        offsets_skin1 = []
        bondlist2 = []
        offsets_skin2 = []
        speciesarray = np.array(atoms.info['species'])
        for b in self:
            if b.search == 1:
                temp = bondlists0[(speciesarray[bondlists0[:, 0]] == b.symbol1) 
                              & (speciesarray[bondlists0[:, 1]] == b.symbol2)]
                bondlist1.extend(temp)
                temp = offsets_skin0[(speciesarray[offsets_skin0[:, 0]] == b.symbol1)]
                offsets_skin1.extend(temp)
            elif b.search == 2:
                #  recursively if either sp1 or sp2
                temp = bondlists0[((speciesarray[bondlists0[:, 0]] == b.symbol1) 
                                  & (speciesarray[bondlists0[:, 1]] == b.symbol2))]
                bondlist2.extend(temp)
                temp1 = temp.copy()
                temp1[:, 0] = temp[:, 1]
                temp1[:, 1] = temp[:, 0]
                temp1[:, 2:] = -temp[:, 2:]
                bondlist2.extend(temp1)
                temp = offsets_skin0[(speciesarray[offsets_skin0[:, 0]] == b.symbol1)
                                   | (speciesarray[offsets_skin0[:, 0]] == b.symbol2)]
                offsets_skin2.extend(temp)
        return np.array(offsets_skin1), np.array(bondlist1), np.array(offsets_skin2), np.array(bondlist2)


def get_bondtable(speciesdict, cutoff = 1.3):
    """
    """
    from batoms.data import default_bonds
    bondtable = {}
    for species1 in speciesdict:
        element1 = species1.split('_')[0]
        color1 = speciesdict[species1]['color']
        radius1 = cutoff * speciesdict[species1]['radius']
        for species2 in speciesdict:
            element2 = species2.split('_')[0]
            pair = (element1, element2)
            if pair not in default_bonds: continue
            color2 = speciesdict[species2]['color']
            radius2 = cutoff * speciesdict[species2]['radius']
            bondmax = radius1 + radius2
            bondtable[(species1, species2)] = [0.5, bondmax, default_bonds[pair][0], 
                            default_bonds[pair][1], color1, color2, 0.10, '1']
    # special for hydrogen bond
    if ('H', 'O') in bondtable:
        bondtable[('H', 'O')] = [1.2, 2.1, 0, False, [0.1, 0.1, 0.1, 1.0], [0.1, 0.1, 0.1, 1.0], 0.03, '2']
    if ('H', 'N') in bondtable:
        bondtable[('H', 'N')] = [1.2, 2.1, 0, False, [0.1, 0.1, 0.1, 1.0], [0.1, 0.1, 0.1, 1.0], 0.03, '2']
    return bondtable

def build_bondlists(atoms, cutoff):
    """
    The default bonds are stored in 'default_bonds'
    Get all pairs of bonding atoms
    remove_bonds
    """
    from batoms.neighborlist import neighbor_list
    if len(cutoff) == 0: return {}
    #
    tstart = time()
    nli, nlj, nlS = neighbor_list('ijS', atoms, cutoff, self_interaction=False)
    bondlists = np.append(np.array([nli, nlj], dtype=int).T, np.array(nlS, dtype=int), axis = 1)
    # print('build_bondlists: {0:10.2f} s'.format(time() - tstart))
    return bondlists

def calc_bond_data(batoms, bondlists, bondsetting):
    """
    """
    from ase.data import chemical_symbols
    atoms = batoms.get_atoms_with_boundary()
    positions = atoms.positions
    chemical_symbols = np.array(chemical_symbols)
    if 'species' not in atoms.info:
        atoms.info['species'] = atoms.get_chemical_symbols()
    speciesarray = np.array(atoms.info['species'])
    bond_kinds = {}
    if len(bondlists) == 0:
        bond_kinds = {}
        return
    tstart = time()
    for b in bondsetting:
        spi = b.symbol1
        spj = b.symbol2
        bondlists1 = bondlists[(speciesarray[bondlists[:, 0]] == spi) 
                    & (speciesarray[bondlists[:, 1]] == spj)]
        if len(bondlists1) == 0: continue
        offset = bondlists1[:, 2:5]
        R = np.dot(offset, atoms.cell)
        vec = positions[bondlists1[:, 0]] - (positions[bondlists1[:, 1]] + R)
        length = np.linalg.norm(vec, axis = 1)
        nvec = vec/length[:, None]
        pos = [positions[bondlists1[:, 0]] - nvec*batoms[spi].size*0.5,
                positions[bondlists1[:, 1]] + R + nvec*batoms[spj].size*0.5]
        vec = pos[0] - pos[1]
        length = np.linalg.norm(vec, axis = 1)
        nvec = vec/length[:, None]
        nvec = nvec + 1e-8
        # verts, faces, for instancing
        # v1 = nvec + np.array([1.2323, 0.493749, 0.5604937284])
        # tempv = np.einsum("ij, ij->i", v1, nvec)
        # v11 = v1 - (nvec.T*tempv).T
        # templengh = np.linalg.norm(v11, axis = 1)
        # v11 = v11/templengh[:, None]/2.828427
        # tempv = np.cross(nvec, v11)
        # v22 = (tempv.T*(length*length)).T
        #
        kinds = [('%s_%s'%(spi, spj), b.color1), 
                    ('%s_%s_%s'%(spi, spj, spi), b.color1), 
                    ('%s_%s_%s'%(spi, spj, spj), b.color2),
                ]
        for kind, color in kinds:
            if kind not in bond_kinds:
                bond_kinds[kind] = {'color': color[:3], 'verts': [], 'transmit': color[3], 
                                    'width': b.width,
                                    'centers': [],
                                    'lengths': [],
                                    'normals': [], 
                                    'style': b.style}
        center0 = (pos[0] + pos[1])/2.0
        # Unicolor cylinder
        if b.style == '0':
                bond_kinds[kinds[0][0]]['centers'] = center0
                bond_kinds[kinds[0][0]]['lengths'] = length
                bond_kinds[kinds[0][0]]['normals'] = nvec
        # Bicolor cylinder
        elif b.style == '1':
            length = length/2.0
            for i in range(1, 3):
                center = (center0 + pos[i - 1])/2.0
                bond_kinds[kinds[i][0]]['centers'] = center
                bond_kinds[kinds[i][0]]['lengths'] = length
                bond_kinds[kinds[i][0]]['normals'] = nvec
                # bond_kinds[kinds[i][0]]['verts'] = center + v11
                # bond_kinds[kinds[i][0]]['verts'] = np.append(bond_kinds[kinds[i][0]]['verts'], center - v11, axis = 0)
                # bond_kinds[kinds[i][0]]['verts'] = np.append(bond_kinds[kinds[i][0]]['verts'], center + v22, axis = 0)
                # bond_kinds[kinds[i][0]]['verts'] = np.append(bond_kinds[kinds[i][0]]['verts'], center - v22, axis = 0)
        # Dashed line
        elif b.style == '2':
            length = length/4.0
            for i in range(1, 3):
                bond_kinds[kinds[i][0]]['centers'] = []
                bond_kinds[kinds[i][0]]['lengths'] = []
                bond_kinds[kinds[i][0]]['normals'] = []
                step = 0.2
                maxlength = length.max()
                center = (center0 + pos[i - 1])/2.0
                for d in np.arange(-maxlength, maxlength, step):
                    # offset0 = np.linspace(-2, 2, nc)
                    # np.where(offset0>-length/2.0 & offset<length/2.0)
                    offset = nvec*d
                    center1 = center + offset
                    ind = np.where( (d>-length) & (d<length))[0]
                    bond_kinds[kinds[i][0]]['centers'].extend(center1[ind])
                    bond_kinds[kinds[i][0]]['lengths'].extend([step/2]*len(ind))
                    bond_kinds[kinds[i][0]]['normals'].extend(nvec[ind])
        # Dotted line
        elif b.style == '3':
            length = length/4.0
            for i in range(1, 3):
                bond_kinds[kinds[i][0]]['centers'] = []
                bond_kinds[kinds[i][0]]['lengths'] = []
                bond_kinds[kinds[i][0]]['normals'] = []
                step = 0.05
                maxlength = length.max()
                center = (center0 + pos[i - 1])/2.0
                for d in np.arange(-maxlength, maxlength, step):
                    # offset0 = np.linspace(-2, 2, nc)
                    # np.where(offset0>-length/2.0 & offset<length/2.0)
                    offset = nvec*d
                    center1 = center + offset
                    ind = np.where( (d>-length) & (d<length))[0]
                    bond_kinds[kinds[i][0]]['centers'].extend(center1[ind])
                    bond_kinds[kinds[i][0]]['lengths'].extend([step/4]*len(ind))
                    bond_kinds[kinds[i][0]]['normals'].extend(nvec[ind])
    # print('calc_bond_data: {0:10.2f} s'.format(time() - tstart))
    return bond_kinds





if __name__ == "__main__":
    from ase.build import bulk, molecule
    from ase.atoms import Atoms
    from ase.io import read
    from ase.visualize import view
    from ase.neighborlist import neighbor_list
    from boundary import search_boundary
    # atoms = bulk('Pt', cubic = True)
    atoms = read('docs/source/_static/datas/tio2.cif')
    # atoms = read('docs/source/_static/datas/mof-5.cif')
    atoms = atoms*[4, 4, 4]
    # positions1, offsets1, positions2, offsets2 = search_boundary(atoms.positions, atoms.cell, boundary=[[-0.6, 1.6], [-0.6, 1.6], [-0.6, 1.6]])
    bondsetting = {('Ti', 'O'): [0, 2.5, True, False], ('O', 'O'): [0, 1.5, False, False]}
    bondlists = build_bondlists(atoms, bondsetting)
    


import os
from ase import io
from batoms import Batoms

def read(filename, **kwargs):
    """
    wrapper function for ase.io.read
    """
    atoms = io.read(filename=filename, **kwargs)
    batoms = Batoms(label = os.path.split(filename)[1], atoms=atoms)
    return batoms



import pytest
from batoms.render import Render
from batoms.batoms import Batoms
from batoms.butils import removeAll
import numpy as np


def test_render():
    """
    """
    removeAll()
    r = Render('test')
    assert isinstance(r, Render)
    #
    from ase.build import fcc111
    removeAll()
    atoms = fcc111('Au', size = (4, 4, 4), vacuum=0)
    au111 = Batoms(label = 'au111', atoms = atoms)
    au111.cell[2, 2] += 5
    au111.draw_cell()
    au111.render.run([1, 1, 1], engine = 'workbench', resolution_x = 200, output='au111')
    au111.render.run([1, 1, 1], engine = 'eevee', resolution_x = 200, output='au111-eevee')
    au111.render.run([1, 1, 1], engine = 'cycles', resolution_x = 200, output='au111-cycles')

test_render()


import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name="batoms",
    version="0.1.0",
    description="Drawing and rendering beautiful atoms, molecules using Blender.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/superstar54/batoms",
    author="Xing Wang",
    author_email="xingwang1991@gmail.com",
    license="GPL",
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        "Programming Language :: Python :: 3.9",
    ],
    packages=["batoms"],
    entry_points={'console_scripts': ['batoms=batoms.cli.main:main']},
    install_requires=["ase", "numpy", "scipy", "scikit-image"],
    python_requires='>=3.9',
)



import pytest
from batoms.butils import removeAll
from batoms.batoms import Batoms
from batoms.bio import read
import numpy as np


def test_replace():
    from ase.build import molecule
    from batoms.batoms import Batoms
    from batoms.butils import removeAll
    removeAll()
    co = Batoms('co', atoms = molecule('CO'))
    co.cell = [2, 2, 3]
    co.pbc = True
    co.repeat([2, 2, 2])
    co.replace('C', 'C_1', [5])
    co.model_type = 1
    
def test_polyhedra():
    from batoms.bio import read
    from batoms.butils import removeAll
    removeAll()
    tio2 = read('../../docs/source/_static/datas/tio2.cif')
    tio2.boundary = 0.01


def test_search_bond():
    from batoms.bio import read
    from batoms.butils import removeAll
    removeAll()
    pk = read('../../docs/source/_static/datas/perovskite.cif')
    pk.repeat([2, 2, 2])
    pk.boundary = 0.01
    pk.model_type = 2
    pk.draw_cell()
    pk.render.run([0, 1, 0], engine = 'eevee', output = 'perovskite.png')


def test_search_bond_2():
    from batoms.bio import read
    from batoms.butils import removeAll
    removeAll()
    mol = read('../../docs/source/_static/datas/anthraquinone.cif')
    mol.boundary = 0.01
    mol.draw_cell()
    mol.model_type = 1
    mol.render.run([1, -0.3, 0.1], engine = 'eevee', output = 'anthraquinone.png')

def test_search_bond_urea():
    from batoms.bio import read
    from batoms.butils import removeAll
    removeAll()
    mol = read('../../docs/source/_static/datas/urea.cif')
    mol.boundary = 2
    mol.draw_cell()
    mol.model_type = 1
    mol.render.run([1, -0.3, 0.1], engine = 'eevee', output = 'urea.png')


def test_search_bond_3():
    """
    Big system
    """
    from batoms.bio import read
    from batoms.butils import removeAll
    removeAll()
    mof = read('../../docs/source/_static/datas/mof-5.cif')
    mof.boundary = 0.01
    mof.bondsetting[('Zn-O')].polyhedra = True
    mof.model_type = 1
    mof.render.run([0, 1, 0], engine = 'eevee', output = 'mof-5.png')




def test_hydrogen_bond():
    from ase.build import molecule
    from batoms.batoms import Batoms
    from batoms.butils import removeAll
    removeAll()
    h2o = molecule('H2O')
    h2o2 = molecule('H2O')
    h2o2.rotate(90, 'x')
    h2o2.translate([0, 0, 3])
    h2o = h2o + h2o2
    h2o = Batoms(label = 'h2o', atoms = h2o)
    h2o.bondsetting['H-O'].min = 2.0
    h2o.bondsetting['H-O'].max = 3.0
    h2o.bondsetting['H-O'].width = 0.01
    h2o.bondsetting['H-O'].style = '2'
    h2o.model_type = 1
    h2o.render.run([1, 0 ,0], engine = 'eevee')



if __name__ == '__main__':
    # test_replace()
    # test_polyhedra()
    # test_hydrogen_bond()
    # test_search_bond()
    # test_search_bond_2()
    test_search_bond_3()
    print('\n Bondsetting: All pass! \n')


"""
"""
import bpy
import numpy as np
from time import time

class Isosurfacesetting():
    """
    Isosurfacesetting object

    The Isosurfacesetting object store the isosurface information.

    Parameters:

    label: str
        The label define the batoms object that a Setting belong to.

    """
    def __init__(self, label, volume) -> None:
        self.label = label
        self.name = 'bisosurface'
        self.volume = volume
        self[1] = [0.002, [1, 1, 0, 0.5]]
    def draw_volume(self, volume):
        """
        Draw unit cell by edge, however, can not be rendered.
        """
        # remove old volume point
        if volume is None: return
        name = "volume_%s"%self.label
        if name in bpy.data.objects:
            bpy.data.objects.remove(bpy.data.objects[name], du_unlink = True)
        shape = volume.shape
        volume = volume.reshape(-1, 1)
        npoint = len(volume)
        dn = 3 - npoint % 3
        verts = np.append(volume, np.zeros((dn, 1)), axis = 0)
        verts = verts.reshape(-1, 3)
        mesh = bpy.data.meshes.new("mesh_%s_volume"%self.label)
        mesh.from_pydata(verts, [], [])  
        mesh.update()
        obj = bpy.data.objects.new(name, mesh)
        obj.data = mesh
        obj.bvolume.is_bvolume = True
        obj.bvolume.shape = shape
        obj.bvolume.npoint = npoint
        bpy.data.collections[self.label].children['%s_volume'%self.label].objects.link(obj)
        obj.hide_set(True)
    @property
    def collection(self):
        return self.get_collection()
    def get_collection(self):
        collection = getattr(bpy.data.collections[self.label], self.name)
        return collection
    @property
    def npoint(self):
        return self.get_npoint()
    def get_npoint(self):
        if "volume_%s"%self.label not in bpy.data.objects:
            return 0
        return bpy.data.objects["volume_%s"%self.label].bvolume.npoint
    @npoint.setter
    def npoint(self, npoint):
        self.set_npoint(npoint)
    def set_npoint(self, npoint):
        bpy.data.objects["volume_%s"%self.label].bvolume.npoint = npoint
    @property
    def mesh(self):
        return self.get_mesh()
    def get_mesh(self):
        mesh = bpy.data.objects["volume_%s"%self.label].data
        return mesh
    @property
    def shape(self):
        return self.get_shape()
    def get_shape(self):
        shape = bpy.data.objects["volume_%s"%self.label].bvolume.shape
        return shape
    @property
    def volume(self):
        return self.get_volume()
    def get_volume(self):
        n = len(self.mesh.vertices)
        volume = np.empty(n*3, dtype=np.float64)
        self.mesh.vertices.foreach_get('co', volume)  
        volume = volume.reshape(-1, 1)
        volume = volume[:self.npoint]
        volume = volume.reshape(self.shape)
        return volume
    @volume.setter
    def volume(self, volume):
        self.draw_volume(volume)
    def __setitem__(self, index, value):
        """
        Add isosurface one by one
        """
        p = self.find(index)
        if p is None:
            p = self.collection.add()
        p.name = str(index)
        if isinstance(value, (int, float)):
            value = [value]
        p.level = value[0]
        if len(value) == 2:
            p.color = value[1]
    def set_default(self):
        """
        """
        for sp, data in self.species.items():
            self[sp] = [np.append(data['color'][:3], 0.3), 0.005]
    def add_isosurfaces(self, isosurfacepair):
        for key in isosurfacepair:
            self.set_default(key)
    def remove_isosurfaces(self, isosurfacepair):
        for key in isosurfacepair:
            name = '%s-%s'%(key[0], key[1])
            i = self.collection.find(name)
            if i != -1:
                self.collection.remove(i)
    def __repr__(self) -> str:
        s = '-'*60 + '\n'
        s = 'Center     level                color  \n'
        for iso in self.collection:
            s += '{0:10s}   {1:1.6f}  [{2:1.2f}  {3:1.2f}  {4:1.2f}  {5:1.2f}] \n'.format(\
                iso.name, iso.level, iso.color[0], iso.color[1], iso.color[2], iso.color[3])
        s += '-'*60 + '\n'
        return s
    def __iter__(self):
        item = self.collection
        for i in range(len(item)):
            yield item[i]
    def __len__(self):
        return len(self.collection)
    def find(self, name):
        i = self.collection.find(str(name))
        if i == -1:
            return None
        else:
            return self.collection[i]
    def build_isosurface(self, cell):
        volume = self.volume
        isosurface = []
        for iso in self.collection:
            level = iso.level
            color = iso.color
            verts, faces = calc_isosurface(volume, cell, level)
            isosurface.append((verts, faces, color))
        return isosurface


def calc_isosurface(volume, cell, level,
                    gradient_direction = 'descent',
                    step_size = 1):
    """
    
    Computes an isosurface from a volume grid.
    
    Parameters:
    
    """
    from batoms.tools import get_cell_vertices
    from skimage import measure

    cell_vertices = get_cell_vertices(cell)
    cell_vertices.shape = (2, 2, 2, 3)
    cell_origin = cell_vertices[0,0,0]
    #
    spacing = tuple(1.0/np.array(volume.shape))
    mlevel = np.mean(volume)
    if not level:
        level = mlevel*10
    # print('iso level: {0:1.9f}, iso mean: {1:1.9f}'.format(level, mlevel))
    scaled_verts, faces, normals, values = measure.marching_cubes(volume, level = level,
                    spacing=spacing,gradient_direction=gradient_direction , 
                    allow_degenerate = False, step_size=step_size)
    scaled_verts = scaled_verts.dot(cell)
    scaled_verts -= cell_origin
    faces = list(faces)
    return scaled_verts, faces



import pytest
from batoms.butils import removeAll
from batoms.batoms import Batoms
from batoms.bio import read
import numpy as np



def test_polyhedra():
    from ase.build import molecule
    from batoms.batoms import Batoms
    ch4 = Batoms('ch4', atoms = molecule('CH4'))
    ch4.bondsetting['C-H'].polyhedra = True
    ch4.polyhedrasetting['C'].color = [0.8, 0.1, 0.3, 1.0]
    ch4.model_type = 2


if __name__ == '__main__':
    test_polyhedra()
    print('\n Bondsetting: All pass! \n')


from operator import pos
import numpy as np
from ase import Atoms, Atom, atom, cell
from batoms.data import covalent_radii
from ase.visualize import view
import time
from ase.neighborlist import neighbor_list

    
def default_element_prop(element, color_style = "JMOL"):
    """
    """
    from ase.data import chemical_symbols
    from batoms.data import jmol_colors, cpk_colors, vesta_color
    element_prop = {}
    number = chemical_symbols.index(element)
    if color_style.upper() == 'JMOL':
        color = jmol_colors[number]
    elif color_style.upper() == 'CPK':
        color = cpk_colors[number]
    elif color_style.upper() == 'VESTA':
        color = vesta_color[element]
    radius = covalent_radii[number]
    element_prop['element'] = element
    element_prop['color'] = color
    element_prop['radius'] = radius
    element_prop['transmit'] = 1.0
    return element_prop

def get_atom_kind(element, positions = [], color_style = "JMOL", scale = [1, 1, 1], props = {}):
    """
    """
    atom_kind = default_element_prop(element, color_style = color_style)
    atom_kind['scale'] = scale
    atom_kind['positions'] = positions
    atom_kind['balltype'] = None
    atom_kind.update(props)
    return atom_kind

def get_polyhedra_kind(color, edgewidth = 0.01, props = {}):
    polyhedra_kind = {'color': color[:3], 'transmit': color[3],
                     'edgewidth': edgewidth}
    vertices = []
    edges = []
    faces = []
    polyhedra_kind.update({'vertices': vertices, 'edges': edges, 'faces': faces})
    lengths = []
    centers = []
    normals = []
    polyhedra_kind['edge_cylinder'] = {'lengths': lengths, 'centers': centers, 'normals': normals}
    polyhedra_kind['edge_cylinder']['color'] = (1.0, 1.0, 1.0)
    polyhedra_kind['edge_cylinder']['transmit'] = 1.0
    polyhedra_kind.update(props)
    return polyhedra_kind

def euler_from_vector(normal, s = 'zxy'):
    from scipy.spatial.transform import Rotation as R
    normal = normal/np.linalg.norm(normal)
    vec = np.cross([0.0000014159, 0.000001951, 1], normal)
    vec = vec/np.linalg.norm(vec)
    # print(vec)
    # ang = np.arcsin(np.linalg.norm(vec))
    ang = np.arccos(normal[2])
    vec = -1*ang*vec
    # print(vec)
    r = R.from_rotvec(vec)
    euler = r.as_euler()
    return euler


def getEquidistantPoints(p1, p2, n):
    return zip(np.linspace(p1[0], p2[0], n+1), np.linspace(p1[1], p2[1], n+1), np.linspace(p1[2], p2[2], n+1))

def get_cell_vertices(cell):
    """
    """
    cell = np.array(cell)
    cell = cell.reshape(3, 3)
    cell_vertices = np.empty((2, 2, 2, 3))
    for c1 in range(2):
        for c2 in range(2):
            for c3 in range(2):
                cell_vertices[c1, c2, c3] = np.dot([c1, c2, c3],
                                                    cell)
    cell_vertices.shape = (8, 3)
    return cell_vertices
def get_canvas(atoms, direction = [0, 0 ,1], margin = 1, show_unit_cell = True):
    """
    """
    from scipy.spatial.transform import Rotation as R
    canvas = np.zeros([2, 3])
    positions = atoms.positions
    cell_vertices = get_cell_vertices(atoms.cell)
    for i in range(3):
        canvas[0, i] = positions[:, i].min()
        canvas[1, i] = positions[:, i].max()
        if show_unit_cell:
            canvas[0, i] = min(canvas[0, i], cell_vertices[:, i].min())
            canvas[1, i] = max(canvas[1, i], cell_vertices[:, i].max())
    #
    canvas1 = np.zeros([2, 3])
    direction = np.array(direction)
    nz = direction/np.linalg.norm(direction)
    nx = np.cross([0, 0, 1], nz) + np.array([0.000001, 0, 0])
    nx = nx/np.linalg.norm(nx)
    ny = np.cross(nz, nx) + np.array([0, 0.000001, 0])
    ny = ny/np.linalg.norm(ny)
    #
    projxy = positions.copy()
    projx = np.zeros((len(positions), 1))
    projy = np.zeros((len(positions), 1))
    projz = np.dot(positions, nz)
    for i in range(len(positions)):
        projxy[i] = positions[i] - projz[i]*nz
        projx[i] = np.dot(projxy[i], nx)
        projy[i] = np.dot(projxy[i], ny)
    #
    projcellxy = cell_vertices.copy()
    projcellx = np.zeros((len(cell_vertices), 1))
    projcelly = np.zeros((len(cell_vertices), 1))
    for i in range(len(cell_vertices)):
        projcellxy[i] = cell_vertices[i] - np.dot(cell_vertices[i], nz)*nz
        projcellx[i] = np.dot(projcellxy[i], nx)
        projcelly[i] = np.dot(projcellxy[i], ny)
    canvas1[0, 0] = projx.min()
    canvas1[1, 0] = projx.max()
    canvas1[0, 1] = projy.min()
    canvas1[1, 1] = projy.max()
    if show_unit_cell:
        canvas1[0, 0] = min(canvas1[0, 0], projcellx.min())
        canvas1[1, 0] = max(canvas1[1, 0], projcellx.max())
        canvas1[0, 1] = min(canvas1[0, 1], projcelly.min())
        canvas1[1, 1] = max(canvas1[1, 1], projcelly.max())
    canvas1[0, :] -= margin
    canvas1[1, :] += margin
    canvas1[0, 2] = 0
    canvas1[1, 2] = projz.max()
    return canvas, canvas1
def find_cage_sphere(cell, positions, radius, step = 1.0, boundary = [[0, 1], [0, 1], [0, 1]]):
    from ase.cell import Cell
    from scipy.spatial import distance

    cell = Cell(cell)
    a, b, c, alpha, beta, gamma = cell.cellpar()
    na, nb, nc = int(a/step),int(b/step), int(c/step)
    x = np.linspace(boundary[0][0], boundary[0][1], na)
    y = np.linspace(boundary[1][0], boundary[1][1], nb)
    z = np.linspace(boundary[2][0], boundary[2][1], nc)
    positions_v = np.vstack(np.meshgrid(x, y, z)).reshape(3,-1).T
    positions_v = np.dot(positions_v, cell)
    dists = distance.cdist(positions_v, positions)
    # dists<3.0
    flag = np.min(dists, axis = 1) >radius
    return positions_v[flag]
def find_cage(cell, positions, radius, step = 1.0, boundary = [[0, 1], [0, 1], [0, 1]]):
    from ase.cell import Cell
    from scipy.spatial import distance

    cell = Cell(cell)
    a, b, c, alpha, beta, gamma = cell.cellpar()
    na, nb, nc = int(a/step),int(b/step), int(c/step)
    x = np.linspace(boundary[0][0], boundary[0][1], na)
    y = np.linspace(boundary[1][0], boundary[1][1], nb)
    z = np.linspace(boundary[2][0], boundary[2][1], nc)
    positions_v = np.vstack(np.meshgrid(x, y, z)).reshape(3,-1).T
    positions_v = np.dot(positions_v, cell)
    dists = distance.cdist(positions_v, positions)
    # dists<3.0
    flag = np.min(dists, axis = 1) >radius
    return positions_v[flag]

def rotation_matrix_from_vectors(vec1, vec2):

    a, b = (vec1 / np.linalg.norm(vec1)).reshape(3), (vec2 / np.linalg.norm(vec2)).reshape(3)
    v = np.cross(a, b)
    c = np.dot(a, b)
    s = np.linalg.norm(v)
    kmat = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
    rotation_matrix = np.eye(3) + kmat + kmat.dot(kmat) * ((1 - c) / (s ** 2))
    return rotation_matrix




if __name__ == "__main__":
    from ase.build import bulk
    from ase.atoms import Atoms
    from ase.io import read
    from ase.visualize import view
    from ase.neighborlist import neighbor_list
    # atoms = bulk('Pt') #, cubic = True)
    atoms = bulk('Pt', cubic = True)
    # atoms = atoms*[2, 2, 1]
    # atoms = read('docs/source/_static/datas/tio2.cif')
    # atoms = read('docs/source/_static/datas/mof-5.cif')
    # nli_min, nlj_min, nlS_min = neighbor_list('ijS', atoms, cutoff={('Ti', 'O'): 2.5, ('O', 'O'): 1.4}, self_interaction=False)
    # print(nli_min)
    # print(nlj_min)
    # view(atoms)
    # positions = find_cage(atoms.cell, atoms.positions, 9.0, step = 1.0)
    # vatoms = Atoms(['Au']*len(positions), positions=positions)
    